# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse
import sys

from hio import help
from hio.base import doing

from keri.app import signing
from keri.app.cli.common import existing
from keri.core import serdering
from keri.vdr import credentialing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='List credentials and check mailboxes for any newly issued credentials')
parser.set_defaults(handler=lambda args: export_credentials(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

parser.add_argument("--said", "-s", help="SAID of the credential to export.", required=False, default=None)
parser.add_argument("--tels", help="export the transaction event logs for the credential and any chained credentials",
                    action="store_true")
parser.add_argument("--kels", help="export the key event logs for the issuer's of the credentials", action="store_true")
parser.add_argument("--chains", help="export any chained credentials", action="store_true")
parser.add_argument("--full", help="export credential, signatures, tels, kels and full chains", action="store_true")
parser.add_argument("--files", help="export artifacts to individual files keyed off of AIDs or SAIDS, default is "
                                    "stdout", action="store_true")


def export_credentials(args):
    """ Command line list credential registries handler

    """
    tels = args.tels
    kels = args.kels
    chains = args.chains if args.chains is not None else {}

    if args.full:
        tels = kels = chains = True

    ed = ExportDoer(name=args.name,
                    alias=args.alias,
                    base=args.base,
                    bran=args.bran,
                    said=args.said,
                    tels=tels,
                    kels=kels,
                    chains=chains,
                    files=args.files)
    return [ed]


class ExportDoer(doing.DoDoer):

    def __init__(self, name, alias, base, bran, said, tels, kels, chains, files):
        self.said = said
        self.tels = tels
        self.kels = kels
        self.chains = chains
        self.files = files

        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        self.rgy = credentialing.Regery(hby=self.hby, name=name, base=base)

        doers = [doing.doify(self.exportDo)]

        super(ExportDoer, self).__init__(doers=doers)

    def exportDo(self, tymth, tock=0.0):
        """ Export credential from store and any related material

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if self.said is None:
            for (said,), _ in self.rgy.reger.creds.getItemIter():
                self.outputCred(said=said)

        else:
            self.outputCred(said=self.said)

    def outputCred(self, said):
        creder, *_ = self.rgy.reger.cloneCred(said=said)

        if self.kels:
            issr = creder.issuer
            self.outputKEL(issr)

        if self.tels:
            if creder.regi is not None:
                self.outputTEL(creder.regi)
                self.outputTEL(creder.said)

        if self.chains:
            chains = creder.edge if creder.edge is not None else {}
            saids = []
            for key, source in chains.items():
                if key == 'd':
                    continue

                if not isinstance(source, dict):
                    continue

                saids.append(source['n'])

            for said in saids:
                self.outputCred(said)

        (prefixer, seqner, saider) = self.rgy.reger.cancs.get(keys=(creder.said,))
        if self.files:
            f = open(f"{creder.said}-acdc.cesr", 'w')
            f.write(signing.serialize(creder, prefixer, seqner, saider))
            f.close()
        else:
            sys.stdout.write(signing.serialize(creder, prefixer, seqner, saider).decode("utf-8"))
            sys.stdout.flush()

    def outputTEL(self, regk):
        f = None
        if self.files:
            f = open(f"{regk}-tel.cesr", "w")

        for msg in self.rgy.reger.clonePreIter(pre=regk):
            if f is not None:
                f.write(msg.decode("utf-8"))
            else:
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                sys.stdout.write(serder.raw.decode("utf-8"))
                sys.stdout.write(atc.decode("utf-8"))

        if f is not None:
            f.close()

    def outputKEL(self, pre):
        f = None
        if self.files:
            f = open(f"{pre}-kel.cesr", "w")

        for msg in self.hby.db.clonePreIter(pre=pre):
            if f is not None:
                f.write(msg.decode("utf-8"))
            else:
                serder = serdering.SerderKERI(raw=msg)
                atc = msg[serder.size:]
                sys.stdout.write(serder.raw.decode("utf-8"))
                sys.stdout.write(atc.decode("utf-8"))

        if f is not None:
            f.close()



# -*- encoding: utf-8 -*-
"""
KERI
keri.help package

utility functions

"""
# Setup module global ogler as package logger factory. This must be done on
#  import to ensure global is defined so all modules in package have access to
#  logggers via ogling.ogler.getLoggers(). May always change level and reopen log file
#  if need be

from hio.help import ogling

#  want help.ogler always defined by default
ogler = ogling.initOgler(prefix='keri', syslogged=False)  # inits once only on first import

from .helping import nowIso8601, toIso8601, fromIso8601



# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri.app.cli.common import existing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Display OOBIs waiting for resolution and allow for clean up')
parser.set_defaults(handler=lambda args: list_oobis(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

def list_oobis(args):
    """ Command line list handler

    """
    kwa = dict(args=args)
    return [doing.doify(oobis, **kwa)]


def oobis(tymth, tock=0.0, **opts):
    _ = (yield tock)

    args = opts["args"]
    name = args.name
    base = args.base
    bran = args.bran


    with existing.existingHby(name=name, base=base, bran=bran) as hby:
        oobis = []
        for (oobi,), _ in hby.db.oobis.getItemIter():
            oobis.append(oobi)

        if not oobis:
            print("No OOBIs waiting resolution")
            return

        print("The following OOBIs are pending resolution (enter number to delete):")
        for idx, oobi in enumerate(oobis):
            print(f"\t{idx+1}: {oobi}")

        try:
            idx = input("Number: ")
            idx = int(idx) - 1
            if 0 <= idx < len(oobis):
                if hby.db.oobis.rem(oobis[idx]):
                    print("OOBI deleted")
                else:
                    print("Unable to delete OOBI")
            else:
                print("Invalid number\n")
        except ValueError:
            print("Invalid number\n")





# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse

from hio.base import doing

from keri.help import helping

parser = argparse.ArgumentParser(description='Print a new time')
parser.set_defaults(handler=lambda args: handler(args))


def handler(_):
    return [doing.doify(time)]


def time(tymth, tock=0.0):
    """ time
    """
    _ = (yield tock)

    print(helping.nowIso8601())



# -*- encoding: utf-8 -*-
"""
keri.kli.common.terming module

"""


class Colors:
    """ Terminal colors """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Symbols:
    """ Terminal Symbols """
    CHECKMARK = '\u2714'
    FAILED = '\u2718'



# -*- encoding: utf-8 -*-
"""
keri.kli.commands module

"""
import argparse
import random
import sys
import time
from collections import namedtuple

from hio import help
from hio.base import doing
from keri.app import agenting, indirecting, habbing, forwarding
from keri.app.cli.common import existing, terming
from keri.app.habbing import GroupHab
from keri.core import coring

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Perform a one time watch of all current local AIDs')
parser.set_defaults(handler=lambda args: watch(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='Human readable reference', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")

# Authentication for keystore
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument('--aeid', help='qualified base64 of non-transferable identifier prefix for  authentication '
                                   'and encryption of secrets in keystore', default=None)

Stateage = namedtuple("Stateage", 'even ahead behind duplicitous')

States = Stateage(even="even", ahead="ahead", behind="behind", duplicitous="duplicitous")


class WitnessState:
    wit: str
    state: Stateage
    sn: int
    dig: str


def watch(args):
    name = args.name

    qryDoer = WatchDoer(name=name, base=args.base, bran=args.bran)
    return [qryDoer]


class WatchDoer(doing.DoDoer):

    def __init__(self, name, base, bran, **kwa):
        doers = []
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hbyDoer = habbing.HaberyDoer(habery=self.hby)  # setup doer
        self.cues = help.decking.Deck()

        self.mbd = indirecting.MailboxDirector(hby=self.hby, topics=["/replay", "/receipt", "/reply"])
        self.postman = forwarding.Poster(hby=self.hby)
        doers.extend([self.hbyDoer, self.mbd, self.postman, doing.doify(self.cueDo)])

        self.toRemove = list(doers)
        doers.extend([doing.doify(self.watchDo)])
        super(WatchDoer, self).__init__(doers=doers, **kwa)

    def watchDo(self, tymth, tock=0.0, **opts):
        """
        Returns:  doifiable Doist compatible generator method
        Usage:
            add result of doify on this method to doers list
        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        for hab in self.hby.habs.values():
            print(f"Processing {hab.name} ({hab.pre})")
            if len(hab.kever.wits) == 0:
                print("\tNo witnesses, skipping.")
                continue

            states = []
            for wit in hab.kever.wits:
                sys.stdout.write(f"Checking witness {wit}...")
                sys.stdout.flush()
                keys = (hab.pre, wit)

                # Check for Key State from this Witness and remove if exists
                saider = hab.db.knas.get(keys)
                if saider is not None:
                    hab.db.knas.rem(keys)
                    hab.db.ksns.rem((saider.qb64,))
                    hab.db.ksns.rem((saider.qb64,))

                witer = agenting.messenger(hab, wit)
                self.extend([witer])

                msg = hab.query(pre=hab.pre, src=wit, route="ksn")
                witer.msgs.append(bytearray(msg))

                start = time.perf_counter()
                while not witer.idle:
                    end = time.perf_counter()
                    if end - start > 10:
                        break

                    yield self.tock

                self.remove([witer])

                start = time.perf_counter()
                skip = False
                while True:
                    if (saider := hab.db.knas.get(keys)) is not None:
                        break

                    end = time.perf_counter()
                    if end - start > 10:
                        sys.stdout.write(f"{terming.Colors.FAIL} no response received{terming.Colors.ENDC}")
                        sys.stdout.flush()
                        skip = True
                        break

                    yield self.tock

                print("")

                if skip:
                    continue

                mystate = hab.kever.state()
                witstate = hab.db.ksns.get((saider.qb64,))

                states.append(self.diffState(wit, mystate, witstate))

            # First check for any duplicity, if so get out of here
            dups = [state for state in states if state.state == States.duplicitous]
            ahds = [state for state in states if state.state == States.ahead]
            bhds = [state for state in states if state.state == States.behind]
            if len(dups) > 0:
                print("The following witnesses have a duplicitous event:")
                for state in dups:
                    print(f"\tWitness {state.wit} at Seq No. {state.sn} with digest: {state.dig}")
                print("Further action must be taken to recover from the duplicity")

            elif len(ahds) > 0:
                # Only group habs can be behind their witnesses
                if not isinstance(hab, GroupHab):
                    print("ERROR: Single sig AID behind witnesses, aborting for this AID")
                    continue

                # First check for duplicity among the witnesses that are ahead (possible only if toad is below
                # super majority)
                digs = set([state.dig for state in ahds])
                if len(digs) > 1:  # Duplicity across witness sets
                    print(f"There are multiple duplicitous events on witnesses for {hab.pre}")
                    print("We recommend you abandon this AID")
                else:  # all witnesses that are ahead agree on the event
                    print("The following witnesses have an event that is ahead of the local KEL:")
                    for state in ahds:
                        print(f"\tWitness {state.wit} at Seq No. {state.sn} with digest: {state.dig}")

                state = random.choice(ahds)
                print("If and only if you were expecting to locally be behind your witnesses (multisig for example)")
                print("the following command can be used to locally catch up with your witness:")
                print(f"\n\tkli multisig update --name {self.hby.name} --alias {hab.name} --wit {state.wit} --sn "
                      f"{state.sn} --said {state.dig}\n")

                if len(bhds) > 0:
                    print("You have some witnesses that are also behind you, catch them up afterwards with:")
                    print(f"\n\tkli submit --name {self.hby.name} --alias {hab.name}\n")

            elif len(bhds) > 0:
                print("The following witnesses are behind the local KEL:")
                for state in bhds:
                    print(f"\tWitness {state.wit} at Seq No. {state.sn} with digest: {state.dig}")

                print("Recommend the following command to catch up witnesses:")
                print(f"\n\tkli submit --name {self.hby.name} --alias {hab.name}\n")

            else:
                print(f"Local key state is consistent with the {len(states)} (out of "
                      f"{len(hab.kever.wits)} total) witnesses that responded")

            print()

        self.remove(self.toRemove)

        return

    def cueDo(self, tymth, tock=0.0, **opts):
        """

        Handle cues coming out of our external Mailbox listener and forward to controller
        mailbox if appropriate

        """
        self.wind(tymth)
        self.tock = tock
        yield self.tock

        while True:
            while self.mbd.kvy.cues:
                cue = self.mbd.kvy.cues.popleft()
                cueKin = cue["kin"]  # type or kind of cue
                if cueKin == "keyStateSaved":
                    pass
                    # print("KSN Saved")

                yield self.tock
            yield self.tock

    @staticmethod
    def diffState(wit, preksn, witksn):

        witstate = WitnessState()
        witstate.wit = wit
        mysn = int(preksn.s, 16)
        mydig = preksn.d
        witstate.sn = int(witksn.f, 16)
        witstate.dig = witksn.d

        # At the same sequence number, check the DIGs
        if mysn == witstate.sn:
            if mydig == witstate.dig:
                witstate.state = States.even
            else:
                witstate.state = States.duplicitous

        # This witness is behind and will need to be caught up.
        elif mysn > witstate.sn:
            witstate.state = States.behind

        # mysn < witstate.sn - We are behind this witness (multisig or restore situation).
        # Must ensure that controller approves this event or a recovery rotation is needed
        else:
            witstate.state = States.ahead

        return witstate



# -*- encoding: utf-8 -*-
"""
keri.app.signaling module

"""
import datetime
import time

import falcon
from hio.base import doing
from hio.help import decking

from keri.core import coring
from keri.help import helping


def signal(attrs, topic, ckey=None, dt=None):
    """

    Parameters:
        attrs (dict): payload of the notice
        topic (str): routing for recipient of message
        dt(Optional(str, datetime)): iso8601 formatted datetime of notice
        ckey (str): collapse key

    Returns:
        Notice:  Notice instance

    """
    dt = dt if dt is not None else datetime.datetime.now().isoformat()

    if hasattr(dt, "isoformat"):
        dt = dt.isoformat()

    pad = dict(i="",
               dt=dt,
               r=topic,
               a=attrs
               )

    return Signal(pad=pad, ckey=ckey)


class Signal(coring.Dicter):
    def __init__(self, pad, ckey=None):
        """ New Signal

        Signals with a collapse key will replace any existing signal not yet read with a matching value
        as collapse key

        Parameters:
            pad (dict):  Attribute values that make up the payload of the signal
            ckey (str): The collapse key to use for
        """
        super(Signal, self).__init__(pad=pad)
        self._ckey = ckey

        if 'dt' not in self.pad:
            self.pad['dt'] = helping.nowIso8601()

    @property
    def topic(self):
        if 'r' in self.pad:
            return self.pad['r']
        else:
            return None

    @property
    def ckey(self):
        return self._ckey

    @property
    def dt(self):
        return self.pad['dt']

    @dt.setter
    def dt(self, dt):
        if hasattr(dt, "isoformat"):
            dt = dt.isoformat()

        self.pad['dt'] = dt

    @property
    def attrs(self):
        if 'a' in self.pad:
            return self.pad['a']
        return None


class Signaler(doing.DoDoer):
    """ Class for sending signals to the controller of an agent.

    The signals are just pings to reload data and not persistent messages that can be reread

    """

    SignalTimeout = datetime.timedelta(minutes=10)

    def __init__(self, signals=None):
        """

        Parameters:
        """
        self.signals = signals if signals is not None else decking.Deck()
        doers = [doing.doify(self.expireDo)]
        super(Signaler, self).__init__(doers=doers)

    def push(self, attrs, topic, ckey=None, dt=None):
        """

        Parameters:
            attrs (dict): signal attributes to push to the cue
            topic (str): routing for recipient of message
            ckey (str): collapse key
            dt(Optional(str, datetime)): iso8601 formatted datetime of notice

        Returns:

        """
        dt = dt if dt is not None else datetime.datetime.now()
        sig = signal(attrs=attrs, topic=topic, ckey=ckey, dt=dt)

        if sig.ckey is not None:
            for i, s in enumerate(self.signals):
                if s.ckey == sig.ckey:
                    self.signals[i] = sig
                    return

        self.signals.append(sig)

    def expireDo(self, tymth=None, tock=0.0):
        """
        Returns doifiable Doist compatible generator method (doer dog)

        Usage:
            add result of doify on this method to doers list

        Parameters:
            tymth is injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock is injected initial tock value

        """
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        while True:  # loop checking for expired messages
            now = datetime.datetime.now()
            toRemove = []
            for sig in self.signals:
                if now - helping.fromIso8601(sig.dt) > self.SignalTimeout:  # Expire messages that are too old
                    toRemove.append(sig)
                yield self.tock

            for sig in toRemove:
                self.signals.remove(sig)

            yield self.tock


def loadEnds(app, *, signals=None):
    """ Load endpoints for agent to controller messages

    Args:
        app (falcon.App): falcon.App to register handlers with:
        signals (Deck): messages for the mailbox stream

    Returns:

    """
    sigEnd = SignalsEnd(signals=signals)
    app.add_route("/mbx", sigEnd)
    return sigEnd


class SignalsEnd:
    """
    HTTP handler that accepts and KERI events POSTed as the body of a request with all attachments to
    the message as a CESR attachment HTTP header.  KEL Messages are processed and added to the database
    of the provided Habitat.

    This also handles `req`, `exn` and `tel` messages that respond with a KEL replay.
    """

    def __init__(self, signals=None):
        """
        Create the MBX HTTP server from the Habitat with an optional Falcon App to
        register the routes with.

        Parameters
             rxbs (bytearray): output queue of bytes for message processing
             mbx (Mailboxer): Mailbox storage
             qrycues (Deck): inbound qry response queues

        """
        self.signals = signals if signals is not None else decking.Deck()

    def on_post(self, req, rep):
        """
        Handles POST for KERI mailbox service.

        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response

        ---
        summary:  Stream Server-Sent Events for KERI mailbox for identifier
        description:  Stream Server-Sent Events for KERI mailbox for identifier
        tags:
           - Mailbox

        responses:
           200:
              content:
                 text/event-stream:
                    schema:
                       type: object
              description: Signal query response for server sent events
           204:
              description: KEL or EXN event accepted.
        """
        rep.set_header('Cache-Control', "no-cache")
        rep.set_header('connection', "close")

        rep.set_header('Content-Type', "text/event-stream")
        rep.status = falcon.HTTP_200
        rep.stream = SignalIterable(signals=self.signals)

    def on_get(self, req, rep):
        """
        Handles GET requests as a stream of SSE events
        Parameters:
              req (Request) Falcon HTTP request
              rep (Response) Falcon HTTP response
        ---
        summary:  Stream Server-Sent Events for KERI mailbox for identifier
        description:  Stream Server-Sent Events for KERI mailbox for identifier
        tags:
           - Mailbox
        responses:
           200:
              content:
                 text/event-stream:
                    schema:
                       type: object
              description: Mailbox query response for server sent events
           204:
              description: KEL or EXN event accepted.
        """
        rep.set_header('Cache-Control', "no-cache")
        rep.set_header('connection', "close")
        rep.set_header('Content-Type', "text/event-stream")

        rep.stream = SignalIterable(signals=self.signals)


class SignalIterable:
    TimeoutMBX = 300

    def __init__(self, signals, retry=5000):
        self.signals = signals
        self.retry = retry

    def __iter__(self):
        self.start = self.end = time.perf_counter()
        return self

    def __next__(self):
        if self.end - self.start < self.TimeoutMBX:
            if self.start == self.end:
                self.end = time.perf_counter()
                return bytes(f"retry: {self.retry}\n\n".encode("utf-8"))

            data = bytearray()
            while self.signals:
                sig = self.signals.popleft()
                topic = sig.topic
                if topic is not None:
                    data.extend(bytearray("id: {}\nretry: {}\nevent: {}\ndata: ".format(sig.rid, self.retry,
                                                                                        topic).encode("utf-8")))
                else:
                    data.extend(bytearray("id: {}\nretry: {}\ndata: ".format(sig.id, self.retry).encode(
                        "utf-8")))

                data.extend(sig.raw)
                data.extend(b'\n\n')
            self.end = time.perf_counter()
            return bytes(data)

        raise StopIteration



# -*- encoding: utf-8 -*-
"""
keri.kli.commands.oobi module

"""
import argparse
import random
import urllib
from urllib.parse import urlparse

import sys
from hio import help
from hio.base import doing

from keri import kering
from keri.app.cli.common import existing

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Generate and print an AID\'s DID.')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)

# Parameters for basic structure of database
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--alias', '-a', help='human readable alias for which to generate a DID', default=None)
parser.add_argument("--role", "-r", help="role of oobis to generate", required=False, default="witness")

# Parameters for Manager access
# passcode => bran
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)
parser.add_argument('--url', '-u', help="generate a DID URL instead of a DID", action="store_true")


def handler(args):
    """ command line method for generating oobies

    Parameters:
        args(Namespace): parse args namespace object

    """
    kwa = dict(args=args)
    return [doing.doify(generate, **kwa)]


def generate(tymth, tock=0.0, **opts):
    """ Command line status handler

    """
    _ = (yield tock)
    args = opts["args"]
    name = args.name
    alias = args.alias
    base = args.base
    bran = args.bran
    role = args.role
    didurl = args.url

    with existing.existingHby(name=name, base=base, bran=bran) as hby:
        if alias is None:
            alias = existing.aliasInput(hby)

        hab = hby.habByName(name=alias)
        if not didurl:
            print(f"did:keri:{hab.pre}")

        elif role in (kering.Roles.witness,):
            if not hab.kever.wits:
                print(f"{alias} identifier {hab.pre} does not have any witnesses.")
                sys.exit(-1)

            wit = random.choice(hab.kever.wits)
            urls = hab.fetchUrls(eid=wit, scheme=kering.Schemes.http) or hab.fetchUrls(eid=wit, scheme=kering.Schemes.https)
            if not urls:
                raise kering.ConfigurationError(f"unable to query witness {wit}, no http endpoint")

            url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
            up = urlparse(url)
            enc = urllib.parse.quote_plus(f"{up.scheme}://{up.hostname}:{up.port}/oobi/{hab.pre}/witness")
            print(f"did:keri:{hab.pre}?oobi={enc}")
        elif role in (kering.Roles.controller,):
            urls = hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.http) or hab.fetchUrls(eid=hab.pre, scheme=kering.Schemes.https)
            if not urls:
                print(f"{alias} identifier {hab.pre} does not have any controller endpoints")
                return
            url = urls[kering.Schemes.http] if kering.Schemes.http in urls else urls[kering.Schemes.https]
            up = urlparse(url)
            enc = urllib.parse.quote_plus(f"{up.scheme}://{up.hostname}:{up.port}/oobi/{hab.pre}/controller")
            print(f"did:keri:{hab.pre}?oobi={enc}")



import argparse
import cmd
import fractions
import json
import os

from hio.base import doing
from keri.app import connecting, apping
from keri.app.cli.common import existing
from keri.core import coring

parser = argparse.ArgumentParser(description='Initialize a prefix')
parser.set_defaults(handler=lambda args: handler(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran

parser.add_argument('--static-threshold', '-t', help='Specify this to switch to using one static number as the signing'
                                                     'threshold for the multisig group AID',
                    dest="static", action="store_true")  # passcode => bran


def handler(args):
    """ Command line list handler

    """
    kwa = dict(args=args)
    return [doing.doify(kims, **kwa)]


def kims(tymth, tock=0.0, **opts):
    _ = (yield tock)

    args = opts["args"]
    name = args.name
    base = args.base
    bran = args.bran
    static = args.static

    with existing.existingHby(name=name, base=base, bran=bran) as hby:
        org = connecting.Organizer(hby=hby)
        contacts = org.list()
        MultiSigShell(hby, contacts, static).cmdloop()
        return True


DEFAULT_PROMPT = '(kims) '


class MultiSigShell(cmd.Cmd):
    intro = 'Welcome to the KERI interactive multisig shell.   Type help or ? to list commands.\n'
    prompt = DEFAULT_PROMPT
    add_cmds = ["participant", "witness", "local"]

    def __init__(self, hby, contacts, static=False):
        self.hby = hby
        self.static = static
        self.transferable = []
        self.nontrans = []
        for contact in contacts:
            prefixer = coring.Prefixer(qb64=contact["id"])
            if prefixer.transferable:
                self.transferable.append(contact)
            else:
                self.nontrans.append(contact)

        self.alias = None
        self.wits = []
        self.delegator = None
        self.participants = []
        self.weights = []
        self.toad = None
        self.thold = None
        self.sith = None

        super(MultiSigShell, self).__init__()

    def reset(self):
        self.alias = None
        self.wits = []
        self.delegator = None
        self.participants = []
        self.toad = None
        self.thold = None
        self.sith = None

    # ----- basic multisig commands -----
    def do_show(self, _):
        """ Print the current start of the multisig group configuration file """
        try:
            self.validate()
        except ValueError as e:
            print(f"Invalid configuration: {e.args[0]}")
            return

        print(json.dumps(self.incept, indent=2))

    def do_delegator(self, arg):
        """ Set the delegator to the AID of the provided contact """
        alias = arg.strip()
        aid = None
        for contact in self.transferable:
            if contact["alias"] == alias:
                aid = contact["id"]
        if aid is not None:
            self.delegator = aid
            print(f"{alias} AID {aid} set as delegator")

    def do_threshold(self, arg):
        """ Set the static integer threshold of the AID.  Must be greater than 0 and less than or equal to the number
        of participants """
        if not self.static:
            print("Using fractionally weighted thresholds, invalid command")
            return

        self.thold = arg

    def complete_delegator(self, text, line, start_index, end_index):
        if text:
            return [
                contact["alias"] for contact in self.transferable if contact["alias"].startswith(text)
            ]
        else:
            return [contact["alias"] for contact in self.transferable]

    def do_add(self, arg):
        """ Add [local | participant | witness ] to the current configuration """
        if arg.startswith("local"):
            alias = arg.removeprefix("local").strip()
            aid = None
            for hab in list(self.hby.habs.values()):
                if hab.name == alias:
                    aid = hab.pre
            if aid is not None:
                self.participants.append(aid)
                print(f"{alias} AID {aid} added to participants")
                if not self.static:
                    self.prompt = "\tEnter weight: "

        elif arg.startswith("participant"):
            alias = arg.removeprefix("participant").strip()
            aid = None
            for contact in self.transferable:
                if contact["alias"] == alias:
                    aid = contact["id"]
            if aid is not None:
                self.participants.append(aid)
                print(f"{alias} AID {aid} added to participants")
                if not self.static:
                    self.prompt = "\tEnter weight: "

        elif arg.startswith("witness"):
            alias = arg.removeprefix("witness").strip()
            aid = None
            for contact in self.nontrans:
                if contact["alias"] == alias:
                    aid = contact["id"]
            if aid is not None:
                self.wits.append(aid)
                print(f"{alias} AID {aid} added to witnesses")

        else:
            print(f"unknown add command {arg}")

    def complete_add(self, text, line, start_index, end_index):
        if line.startswith("add local"):
            if text:
                return [
                    hab.name for hab in list(self.hby.habs.values()) if hab.name.startswith(text)
                ]
            else:
                return [hab.name for hab in list(self.hby.habs.values())]
        elif line.startswith("add participant"):
            if text:
                return [
                    contact["alias"] for contact in self.transferable if contact["alias"].startswith(text)
                ]
            else:
                return [contact["alias"] for contact in self.transferable]
        elif line.startswith("add witness"):
            if text:
                return [
                    contact["alias"] for contact in self.nontrans if contact["alias"].startswith(text)
                ]
            else:
                return [contact["alias"] for contact in self.nontrans]
        else:
            if text:
                return [
                    c for c in self.add_cmds if c.startswith(text)
                ]
            else:
                return self.add_cmds

    def do_save(self, arg):
        """ Save multisig configureation to file which can be passed to `kli multisig incept` """

        if arg == "":
            print("file name is required")
            return

        if os.path.exists(arg):
            print(f"{arg} exists, will not overwrite.")
            return

        try:
            self.validate()
        except ValueError as e:
            print(f"Invalid configuration: {e.args[0]}")
            return

        with open(arg, 'w', encoding="utf-8") as f:
            json.dump(self.incept, f, indent=2)

    def do_exit(self, _):
        """ Exit without saving """
        self.close()
        return True

    @property
    def incept(self):
        incept = dict()
        if self.delegator is not None:
            incept["delpre"] = self.delegator

        incept["aids"] = self.participants
        incept["transferable"] = True
        incept["wits"] = self.wits
        incept["toad"] = self.toad or len(self.wits)

        if self.static:
            incept["isith"] = self.thold if self.thold is not None else f"{len(self.participants)}"
        else:
            incept["isith"] = self.weights

        incept["nsith"] = incept.get("isith")

        return incept

    def precmd(self, line):
        if self.prompt == "\tEnter weight: ":
            try:
                if isNotValidFloat(line):
                    print("Float syntax not supported")
                    return ""

                w = fractions.Fraction(line)
                if not 0 <= w <= 1:
                    print("Invalid fraction value")
                    return ""

            except ValueError:
                print("Invalid fraction value")
                return ""

            self.weights.append(str(w))
            self.prompt = DEFAULT_PROMPT
            return ""

        return line

    def emptyline(self) -> bool:
        return False

    @staticmethod
    def close():
        print("Closing")

    def validate(self):
        if len(self.participants) < 2:
            raise ValueError("not enough participants")

        local = 0
        for p in self.participants:
            if p in self.hby.habs:
                local += 1

        if local < 1:
            raise ValueError("you must have at least one local AID")

        if len(self.wits) < 1:
            raise ValueError("You need at least 1 witness")


def parse(arg):
    """ Convert a series of zero or more numbers to an argument tuple
    Args:
        arg:

    Returns:

    """
    return tuple(map(int, arg.split()))


def isNotValidFloat(s):
    try:
        f = float(s)
        return f not in (0, 1)
    except ValueError:
        return False



import copy

import falcon
from apispec import yaml_utils
from apispec.core import VALID_METHODS, APISpec


class SpecResource:
    """
    Resource for OpenAPI spec

    """

    def __init__(self, app, title, resources, version='1.0.0', openapiVersion="3.0.2"):
        self.spec = APISpec(
            title=title,
            version=version,
            openapi_version=openapiVersion,
        )

        valid_methods = self._get_valid_methods(self.spec)
        routes_to_check = copy.copy(app._router._roots)

        for route in routes_to_check:
            if route.resource is not None:
                operations = dict()
                operations.update(yaml_utils.load_operations_from_docstring(route.resource.__doc__) or {})

                if route.method_map:
                    for method_name, method_handler in route.method_map.items():
                        if method_handler.__module__ == "falcon.responders":
                            continue
                        if method_name.lower() not in valid_methods:
                            continue
                        docstring_yaml = yaml_utils.load_yaml_from_docstring(method_handler.__doc__)
                        operations[method_name.lower()] = docstring_yaml or dict()

                self.spec.path(path=route.uri_template, operations=operations)
            routes_to_check.extend(route.children)

    def _get_valid_methods(self, spec):
        return set(VALID_METHODS[spec.openapi_version.major])

    def on_get(self, _, rep):
        """
        GET endpoint for OpenAPI spec

        Args:
            _: falcon.Request HTTP request
            rep: falcon.Response HTTP response


        """
        rep.status = falcon.HTTP_200
        rep.content_type = "application/yaml"
        rep.data = self.spec.to_yaml().encode("utf-8")
